# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import support

from selene.core.configuration import (
    Config as _CustomConfigForCustomBrowser,
)

Config = _CustomConfigForCustomBrowser

from selene.core.entity import (
    Browser as _CustomBrowser,
)

Browser = _CustomBrowser

"""
Given::

    from selenium.webdriver import Chrome

AND chromedriver executable available in $PATH

WHEN::

    from selene import Browser, Config

    browser = Browser(Config(
        driver=Chrome(),
        base_url='https://google.com',
        timeout=2))

AND (uncomment if needed)::

    # import atexit
    # atexit.register(browser.quit)

AND::

    browser.open('/ncr')

"""

from selene import _managed  # noqa

browser = _managed.browser

from selene.support import by as _by_style_selectors  # noqa

by = _by_style_selectors
"""
AND::

    # browser.element('//*[@name="q"]')).type('selenium').press_enter()
    # OR...

    # browser.element('[name=q]')).type('selenium').press_enter()
    # OR...

    from selene import by

    # browser.element(by.name('q')).type('selenium').press_enter()
    # OR...for total readability

    query = browser.element(by.name('q'))  # actual search doesn't start here, the element is "lazy"
         # here the actual webelement is found
    query.type('selenium').press_enter()
                          # and here it's located again, i.e. the element is "dynamic"

"""

from selene.support.conditions import be as _be_style_conditions  # noqa

be = _be_style_conditions
"""
AND (in case we need to filter collection of items by some condition like visibility)::

    from selene import be

    results = browser.all('.srg .g').filtered_by(be.visible)
"""

from selene.support.conditions import have as _have_style_conditions  # noqa

have = _have_style_conditions
"""
THEN::
    from selene import have

    # results.should(have.size(10))
    # results.first.should(have.text('Selenium automates browsers'))
    # OR...

    results.should(have.size(10))\
        .first.should(have.text('Selenium automates browsers'))

FINALLY (if not registered "atexit" before)::

    browser.quit()
"""

####################
# Advanced Helpers #
####################

from selene.core import command as _advanced_commands  # noqa

command = _advanced_commands
"""
Sometimes you might need some extra actions on elements,
e.g. for workaround something through js::

    from selene import command

    browser.element('#not-in-view').perform(command.js.scroll_into_view)
"""

from selene.core import query as _advanced_queries  # noqa

query = _advanced_queries
# its = _advanced_queries  # TODO: do we really need it too? for better readability: .get(its.text)
"""
Probably you think that will need something like::

    from selene import query

    product_text = browser.element('#to-assert-something-non-standard').get(query.text)
    price = my_int_from(product_text)
    assert price > 100

But usually it's either better to implement your custom condition::

    browser.element('#to-assert-something-non-standard').should(have_in_text_the_int_number_more_than(100))

Where the ``have_in_text_the_int_number_more_than`` is your defined custom condition.
Such condition-based alternative will be less fragile,
because python's assert does not have "implicit waiting",
like selene's should ;)


Furthermore, the good test is when you totally control your test data, and instead::


    product = browser.element('#to-remember-for-future')

    product_text_before = product.get(query.text)
    price_before = my_int_from(product_text_before)

    # ... do something

    product_text_after = product.get(query.text)
    price_after = my_int_from(product_text_after)

    assert price_after > price_before


Normally, better would be to refactor to something like::

    product = browser.element('#to-remember-for-future')

    product.should(have.text('100$'))

    # ... do something

    product.should(have.text('125$'))


You might think you need something like::

    from selene import query

    if browser.element('#i-might-say-yes-or-no').get(query.text) == 'yes':
        # do something...

Or::

    from selene import query

    if browser.all('.option').get(query.size) >= 2:
        # do something...


Maybe one day, you really find a use case:) But for above cases, probably easier would be::

    if browser.element('#i-might-say-yes-or-no').wait_until(have.text('yes'):
        # do something

    # ...

    if browser.all('.i-will-appear').wait_until(have.size_greater_than_or_equal(2)):
        # do something

Or, by using non-waiting versions, if "you are in a rush:)"::

    if browser.element('#i-might-say-yes-or-no').matching(have.text('yes'):
        # do something

    # ...

    if browser.all('.i-will-appear').matching(have.size_greater_than_or_equal(2)):
        # do something

"""

# """
# Just types...
# """
from selene.core.entity import Element, Collection  # noqa

__version__ = '2.0.0rc3post1'



# MIT License
#
# Copyright (c) 2015-2023 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import itertools
from typing import Callable, Optional, Any, Union

from selenium.webdriver.common.options import BaseOptions
from selenium.webdriver.common.service import Service
from selenium.webdriver.remote.webdriver import WebDriver

from selene.common.fp import F
from selene.core.wait import Wait, E

class _DriverStrategiesExecutor:
    def __init__(self, config: Config): ...
    @property
    def driver_instance(self) -> Union[Optional[WebDriver], ...]: ...  # type: ignore
    @property
    def is_driver_managed(self) -> bool: ...
    def build_driver(self) -> WebDriver: ...
    @property
    def is_driver_set(self) -> bool: ...
    @property
    def is_driver_alive(self) -> bool: ...
    @property
    def teardown(self) -> Callable[[WebDriver], None]: ...
    def schedule_teardown(self, get_driver: Callable[[], WebDriver]) -> None: ...
    def get_url(self, url: Optional[str] = None) -> None: ...
    def save_screenshot(self, path: Optional[str] = None) -> Any: ...
    def save_page_source(self, path: Optional[str] = None) -> Any: ...

class Config:
    """
    all options prefixed with _ are so marked as experimental and may change in future
    """

    # Options to customize default driver lifecycle
    driver_name: Optional[str] = None
    driver_options: Optional[BaseOptions] = None
    driver_service: Optional[Service] = None
    driver_remote_url: Optional[str] = None
    hold_driver_at_exit: bool = False
    _reset_not_alive_driver_on_get_url: bool = True
    rebuild_not_alive_driver: bool = False
    _driver_get_url_strategy: Callable[[Config], Callable[[Optional[str]], None]] = ...
    # Options to customize driver management
    build_driver_strategy: Callable[[Config], WebDriver] = ...
    _schedule_driver_teardown_strategy: Callable[
        [Config, Callable[[], WebDriver]], None
    ] = ...
    _teardown_driver_strategy: Callable[[Config, WebDriver], None] = ...
    _is_driver_set_strategy: Callable[[WebDriver], bool] = ...
    _is_driver_alive_strategy: Callable[[WebDriver], bool] = ...
    # Managed Driver
    driver: WebDriver = ...
    # Options to customize this config creation
    _override_driver_with_all_driver_like_options: bool = True
    # Options to customize general Selene behavior
    # > to customize waiting logic
    timeout: float = 4
    poll_during_waits: int = ...  # currently fake option
    _wait_decorator: Callable[[Wait[E]], Callable[[F], F]] = lambda w: lambda f: f
    reports_folder: Optional[str] = ...
    _counter: itertools.count = ...
    save_screenshot_on_failure: bool = True
    save_page_source_on_failure: bool = True
    last_screenshot: Optional[str] = None
    last_page_source: Optional[str] = None
    _save_screenshot_strategy: Callable[[Config, Optional[str]], Any] = ...
    _save_page_source_strategy: Callable[[Config, Optional[str]], Any] = ...
    # Options to customize web browser and elements behavior
    base_url: str = ''
    _get_base_url_on_open_with_no_args: bool = False
    window_width: Optional[int] = None
    window_height: Optional[int] = None
    log_outer_html_on_failure: bool = False
    set_value_by_js: bool = False
    type_by_js: bool = False
    click_by_js: bool = False
    wait_for_no_overlap_found_by_js: bool = False
    # Etc.
    _build_wait_strategy: Callable[[Config], Callable[[E], Wait[E]]] = ...
    _executor: _DriverStrategiesExecutor = ...

    def __init__(
        self,
        *,
        # Options to customize default driver lifecycle
        driver_name: Optional[str] = None,
        driver_options: Optional[BaseOptions] = None,
        driver_service: Optional[Service] = None,
        driver_remote_url: Optional[str] = None,
        hold_driver_at_exit: bool = False,
        _reset_not_alive_driver_on_get_url: bool = True,
        rebuild_not_alive_driver: bool = False,
        _driver_get_url_strategy: Callable[
            [Config], Callable[[Optional[str]], None]
        ] = ...,
        # Options to customize driver management
        build_driver_strategy: Callable[[Config], WebDriver] = ...,
        _schedule_driver_teardown_strategy: Callable[
            [Config, Callable[[], WebDriver]], None
        ] = ...,
        _teardown_driver_strategy: Callable[[Config, WebDriver], None] = ...,
        _is_driver_set_strategy: Callable[[WebDriver], bool] = ...,
        _is_driver_alive_strategy: Callable[[WebDriver], bool] = ...,
        # Managed Driver
        driver: WebDriver = ...,
        # Options to customize this config creation
        _override_driver_with_all_driver_like_options: bool = True,
        # Options to customize general Selene behavior
        # > to customize waiting logic
        timeout: float = 4,
        poll_during_waits: int = ...,  # currently fake option
        _wait_decorator: Callable[[Wait[E]], Callable[[F], F]] = lambda w: lambda f: f,
        reports_folder: Optional[str] = ...,
        _counter: itertools.count = ...,
        save_screenshot_on_failure: bool = True,
        save_page_source_on_failure: bool = True,
        last_screenshot: Optional[str] = None,
        last_page_source: Optional[str] = None,
        _save_screenshot_strategy: Callable[[Config, Optional[str]], Any] = ...,
        _save_page_source_strategy: Callable[[Config, Optional[str]], Any] = ...,
        # Options to customize web browser and elements behavior
        base_url: str = '',
        _get_base_url_on_open_with_no_args: bool = False,
        window_width: Optional[int] = None,
        window_height: Optional[int] = None,
        log_outer_html_on_failure: bool = False,
        set_value_by_js: bool = False,
        type_by_js: bool = False,
        click_by_js: bool = False,
        wait_for_no_overlap_found_by_js: bool = False,
        # Etc.
        _build_wait_strategy: Callable[[Config], Callable[[E], Wait[E]]] = ...,
    ): ...
    def with_(
        self,
        *,
        # Options to customize default driver lifecycle
        driver_name: Optional[str] = None,
        driver_options: Optional[BaseOptions] = None,
        driver_service: Optional[Service] = None,
        driver_remote_url: Optional[str] = None,
        hold_driver_at_exit: bool = False,
        _reset_not_alive_driver_on_get_url: bool = True,
        rebuild_not_alive_driver: bool = False,
        _driver_get_url_strategy: Callable[
            [Config], Callable[[Optional[str]], None]
        ] = ...,
        # Options to customize driver management
        build_driver_strategy: Callable[[Config], WebDriver] = ...,
        _schedule_driver_teardown_strategy: Callable[
            [Config, Callable[[], WebDriver]], None
        ] = ...,
        _teardown_driver_strategy: Callable[[Config, WebDriver], None] = ...,
        _is_driver_set_strategy: Callable[[WebDriver], bool] = ...,
        _is_driver_alive_strategy: Callable[[WebDriver], bool] = ...,
        # Managed Driver
        driver: WebDriver = ...,
        # Options to customize this config creation
        _override_driver_with_all_driver_like_options: bool = True,
        # Options to customize general Selene behavior
        # > to customize waiting logic
        timeout: float = 4,
        poll_during_waits: int = ...,  # currently fake option
        _wait_decorator: Callable[[Wait[E]], Callable[[F], F]] = lambda w: lambda f: f,
        reports_folder: Optional[str] = ...,
        _counter: itertools.count = ...,
        save_screenshot_on_failure: bool = True,
        save_page_source_on_failure: bool = True,
        last_screenshot: Optional[str] = None,
        last_page_source: Optional[str] = None,
        _save_screenshot_strategy: Callable[[Config, Optional[str]], Any] = ...,
        _save_page_source_strategy: Callable[[Config, Optional[str]], Any] = ...,
        # Options to customize web browser and elements behavior
        base_url: str = '',
        _get_base_url_on_open_with_no_args: bool = False,
        window_width: Optional[int] = None,
        window_height: Optional[int] = None,
        log_outer_html_on_failure: bool = False,
        set_value_by_js: bool = False,
        type_by_js: bool = False,
        click_by_js: bool = False,
        wait_for_no_overlap_found_by_js: bool = False,
        # Etc.
        _build_wait_strategy: Callable[[Config], Callable[[E], Wait[E]]] = ...,
    ): ...



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import warnings

from selenium.webdriver.common.by import By


def css(selector):
    return By.CSS_SELECTOR, selector


def xpath(selector):
    return By.XPATH, selector


def id(attribute_value):
    return By.ID, attribute_value


def class_name(value):
    return By.CLASS_NAME, value


def name(attribute_value):
    return By.NAME, attribute_value


def link_text(value):
    return By.LINK_TEXT, value


def partial_link_text(value):
    return By.PARTIAL_LINK_TEXT, value


def _escape_text_quotes_for_xpath(text):
    return 'concat("", "%s")' % (str("\", '\"', \"".join(text.split('"'))))


def text(value):
    return xpath(
        './/*[text()[normalize-space(.) = '
        + _escape_text_quotes_for_xpath(value)
        + ']]'
    )


def partial_text(value):
    return xpath(
        './/*[text()[contains(normalize-space(.), '
        + _escape_text_quotes_for_xpath(value)
        + ')]]'
    )


# TODO: deprecate be_* ? since they hide "xpath" logic, which may not be working in all cases
#       for example in case of mobile...
#       maybe the only good thing to keep is by.text and by.partial_text


def be_following_sibling(with_tag: str = '*'):
    warnings.warn(
        'deprecated; use xpath explicitly to not hide complexity in workaround',
        DeprecationWarning,
    )
    return xpath(f'./following-sibling::{with_tag}')


def be_parent():
    warnings.warn(
        'deprecated; use xpath explicitly to not hide complexity in workaround',
        DeprecationWarning,
    )
    return xpath('..')


def be_first_child(with_tag: str = '*'):
    warnings.warn(
        'deprecated; use xpath explicitly to not hide complexity in workaround',
        DeprecationWarning,
    )
    return xpath(f'./{with_tag}[1]')



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import time
import warnings
from abc import abstractmethod, ABC
from typing import Generic, Callable, TypeVar, Optional

from selene.common import fp
from selene.core.exceptions import TimeoutException

from selene.common.fp import identity

T = TypeVar('T')
R = TypeVar('R')
E = TypeVar('E')
'''
A generic TypeVar to identify an Entity Type, i.e. something to wait on
'''

# TODO: not sure, if we need all these Lambda, Proc, Query, Command in python
# TODO: they was added just to quickly port selenidejs waiting mechanism
# TODO: let's consider removing them... or moving them e.g. to fp

Lambda = Callable[[T], R]
Proc = Callable[[T], None]
Predicate = Callable[[T], bool]


Fn = Callable[[T], R]


# TODO: consider moving outside of "wait" module... because there is no direct cohesion with it
class Query(Generic[E, R]):
    def __init__(self, description: str, fn: Callable[[E], R]):
        self._description = description
        self._fn = fn

    def __call__(self, entity: E) -> R:
        return self._fn(entity)

    def __str__(self):
        return self._description


class Command(Generic[E], Query[E, None]):
    pass


# TODO: provide sexy fluent implementation via builder, i.e. Wait.the(element).atMost(3).orFailWith(hook)
class Wait(Generic[E]):
    # TODO: provide the smallest possible timeout default, something like 1ms
    def __init__(
        self,
        entity: E,
        at_most: float,
        or_fail_with: Optional[Callable[[TimeoutException], Exception]] = None,
        _decorator: Callable[
            [Wait[E]], Callable[[Callable[..., R]], Callable[..., R]]
        ] = lambda _: identity,
    ):
        self.entity = entity
        self._timeout = at_most
        self._hook_failure = or_fail_with or identity
        self._decorator = _decorator

    @property
    def _entity(self):
        warnings.warn(
            'wait.entity will be removed in next version, '
            + 'please use wait.entity instead',
            DeprecationWarning,
        )
        return self.entity

    def at_most(self, timeout: float) -> Wait[E]:
        return Wait(self.entity, timeout, self._hook_failure)

    def or_fail_with(
        self, hook_failure: Optional[Callable[[TimeoutException], Exception]]
    ) -> Wait[E]:
        return Wait(self.entity, self._timeout, hook_failure)

    @property
    def hook_failure(
        self,
    ) -> Optional[Callable[[TimeoutException], Exception]]:
        # TODO: hook_failure or failure_hook?
        return self._hook_failure

    # TODO: consider renaming to `def to(...)`, though will sound awkward when wait.to(condition)
    def for_(self, fn: Callable[[E], R]) -> R:
        def logic(fn: Callable[[E], R]) -> R:
            finish_time = time.time() + self._timeout

            while True:
                try:
                    return fn(self.entity)
                except Exception as reason:
                    if time.time() > finish_time:
                        reason_message = str(reason)

                        reason_string = '{name}: {message}'.format(
                            name=reason.__class__.__name__,
                            message=reason_message,
                        )
                        # TODO: think on how can we improve logging failures in selene, e.g. reverse msg and stacktrace
                        # stacktrace = getattr(reason, 'stacktrace', None)
                        timeout = self._timeout
                        entity = self.entity

                        failure = TimeoutException(
                            f'\n'
                            f'\nTimed out after {timeout}s, while waiting for:'
                            f'\n{entity}.{fn}'
                            f'\n'
                            f'\nReason: {reason_string}'
                        )

                        raise self._hook_failure(failure)

        return self._decorator(self)(logic)(fn)

    def until(self, fn: Callable[[E], R]) -> bool:
        try:
            self.for_(fn)
            return True
        except TimeoutException:
            return False

    # TODO: do we really need these aliases?
    def command(self, description: str, fn: Callable[[E], None]) -> None:
        self.for_(Command(description, fn))

    def query(self, description: str, fn: Callable[[E], R]) -> R:
        return self.for_(Query(description, fn))



# MIT License
#
# Copyright (c) 2023 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from webdriver_manager.chrome import ChromeDriverManager


def _to_find_chromedrivers_from_115(driver_manager: ChromeDriverManager):
    """
    Fixes webdriver_manager issue with latest chrome versions (>= 115.0.5763.0)
    See https://github.com/SergeyPirogov/webdriver_manager/issues/536

    Fix is based on simple ideas from:
    * https://github.com/SergeyPirogov/webdriver_manager/issues/536#issuecomment-1641266654
    * https://github.com/SergeyPirogov/webdriver_manager/issues/536#issuecomment-1641396604

    It monkey patches all driver_manager inner objects (including nested ones)
    and methods

    But also for newer versions of Chrome browsers – it patches PATTERN dict
    from wdm utils.
    This PATTERN patching is the most risky part of the fix,
    because it changes the whole library behavior,
    not just the object passed to this function.
    Let's keep fingers crossed;p
    """

    from webdriver_manager.core import logger as wdm_logger
    from packaging import version
    from webdriver_manager.core import utils as wdm_utils
    from webdriver_manager.core.utils import ChromeType

    # We alias driver object from driver_manager as driver_utils
    # for lesser confusion.
    # In fact wdm driver object is more a bunch of utility functions.
    # From OOP perspective,
    # it's not a pure object representing actual "low level driver binary".
    # And 'driver' name also conflicts with common 'driver' client object from selenium.
    driver_utils = driver_manager.driver
    http_client = driver_utils._http_client

    def chrome_apis_url(endpoint):
        return f'https://googlechromelabs.github.io/chrome-for-testing/{endpoint}'

    good_binary_version = None
    good_binary_url = None
    installed_browser_version = driver_utils.get_browser_version_from_os()

    if installed_browser_version:
        if version.parse(installed_browser_version) < version.parse('115.0.5763.0'):
            # if browser version is less than 115.0.5763.0
            # then we use old wdm logic
            # and just return the driver_manager as is
            return driver_manager

        # patching wdm_utils.PATTERN
        # we need all 4 sub-versions not just 3 of them
        wdm_utils.PATTERN[ChromeType.GOOGLE] = r"\d+\.\d+\.\d+.\d+"
        # retaking version from os after patched pattern
        good_binary_version = wdm_utils.get_browser_version_from_os(
            driver_utils.get_browser_type()
        )
        # let's reset _browser_version to the new 4-sub-versions value
        # from here, we assume that
        # "good" binary version is the same as "good" browser version
        driver_utils._browser_version = good_binary_version

        known_good_versions = (
            http_client.get(chrome_apis_url('known-good-versions-with-downloads.json'))
            .json()
            .get('versions', [])
        )

        matched_version_downloads_chromedriver_per_platform: list = next(
            iter(
                info.get('downloads', {}).get('chromedriver', [])
                for info in known_good_versions
                if info.get('version', None) == good_binary_version
            ),
            [],
        )

        good_binary_url = next(
            iter(
                info.get('url', None)
                for info in matched_version_downloads_chromedriver_per_platform
                if info.get('platform') == driver_utils.get_os_type().replace('_', '-')
            ),
            None,
        )

    if (not installed_browser_version) or (not good_binary_url):
        wdm_logger.log(
            'Failed to get version of Chrome installed at your OS '
            f'(detected os type: {driver_utils._os_type}).'
            f'Going to install the chromedriver binary '
            f'matching latest known stable version of Chrome...'
        )
        last_known_good_versions_with_downloads = http_client.get(
            chrome_apis_url('last-known-good-versions-with-downloads.json')
        ).json()
        stable_channel = last_known_good_versions_with_downloads.get(
            'channels', {}
        ).get('Stable', {})

        # Above and below, we allways use get('key', default) instead of just ['key']
        # to provide default value in case of missing key
        # and make further processing easier and safer
        # (we don't need to check for None when processing via comprehensions).
        # Sometimes, like below we use `get('key', '') or None`
        # instead of just `get('key', None)` or even `get('key)`
        # in order to hint the type of result of the value by the key to get
        # (in this case - string by hinting it via default empty string '').
        # We could use just type hinting on a variable,
        # but we prefer the style consistent with other similar parts of this code...
        last_known_good_version = (stable_channel.get('version', '')) or None
        platform_and_url_pairs = stable_channel.get('downloads', {}).get(
            'chromedriver', []
        )
        url_where_platform_is_os_type = next(
            iter(
                # url is obviously a string and if absent we are interested in None
                # emphasizing this by providing it explicitly
                pair.get('url', None)
                for pair in platform_and_url_pairs
                # .get_os_type() may return None, hence '' as default in get is intended
                if pair.get('platform', '')
                == driver_utils.get_os_type().replace('_', '-')
            ),
            None,
        )
        wdm_logger.log(
            f'latest known stable version of Chrome: {last_known_good_version}'
        )

        good_binary_version = last_known_good_version
        good_binary_url = url_where_platform_is_os_type

    if good_binary_url:
        # it happened that we found good binary url on our own
        # let's monkey patch WDM classes and objects they know it too ;P

        # now we provide exactly correct and ready for download url
        driver_utils._url = good_binary_url

        # let's just return it as is
        driver_utils.get_driver_download_url = lambda: driver_utils._url

        # just in case...
        driver_utils._version = good_binary_version

        # new endpoints provide file for download differently,
        # old wdm filename logic just does not work
        # so let's patch it too
        class PatchedFile(wdm_utils.File):
            filename = good_binary_url.split('/')[-1]  # type: ignore

        # and use PatchedFile on download...
        def download_file(url: str):
            wdm_logger.log(f"About to download new driver from {url}")
            response = http_client.get(url)
            return PatchedFile(response)

        # – by download manager
        # (we safely patch only current object, not the whole class)
        driver_manager._download_manager.download_file = download_file

        # similar story with processing filenames logic...
        def get_binary(files, driver_name):
            try:
                return next(
                    file
                    for file in files
                    if driver_name in file.split('/')[-1].split('.')[0]
                )
            except Exception as e:
                raise Exception(
                    f"Can't find binary for {driver_name} among {files}"
                ) from e

        # safely patching only exactly our manager object's get_binary
        driver_manager.driver_cache._DriverCache__get_binary = get_binary

    return driver_manager
